using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.Text;

namespace ReHasher
{
    internal static class Program
    {
        private static readonly byte[] PublicKey = "537472347762337272795f536d303074686933535f42337374".FromHexString();

        private static readonly string[] CipherText = {
            "15e1c8c4bd2fcb12838d3f48e8fc567a",
            "df4fac1c36648eb270ad899a717dfadf",
            "ccd9b49fc32a726711dd1d7faab7ed81",
        };

        private static void Main(string[] args)
        {
            Console.WriteLine("Hello, world!");
            Console.Write(Encoding.ASCII.GetString(Decrypt(CipherText[0].FromHexString())));
            Console.Write(Encoding.ASCII.GetString(Decrypt(CipherText[1].FromHexString())));
            Console.Write(Encoding.ASCII.GetString(Decrypt(CipherText[2].FromHexString())));
            Console.WriteLine();
        }

        private static byte[] Decrypt(byte[] full)
        {
            var arr1 = new byte[8];
            var arr2 = new byte[8];
            Array.Copy(full, 0, arr1, 0, 8);
            Array.Copy(full, 8, arr2, 0, 8);
            return Decrypt(arr1, arr2);
        }

        private static byte[] Decrypt(byte[] b, byte[] a)
        {
            foreach (byte k in PublicKey.Reverse())
            {
                var c = a.CrippledHash(k);
                byte[] prev_b = a.ToArray();
                b.XorWith(c);
                a = b;
                b = prev_b;
            }

            var newArr = new byte[16];
            a.CopyTo(newArr, 0);
            b.CopyTo(newArr, 8);
            return newArr;
        }

        private static byte[] block1 = new byte[16 / 2];
        private static byte[] block2 = new byte[16 / 2];

        private static byte[] HashBlock(this byte[] block, byte[] key)
        {
            Debug.Assert(block.Length == 16);
            Array.Copy(block, block1, block1.Length);
            Array.Copy(block, block1.Length, block2, 0, block1.Length);

            foreach (byte k in key)
            {
                var newBuffer = new byte[block.Length / 2 + 1];
                block2.CopyTo(newBuffer, 0);
                newBuffer[^1] = k;
                var c = newBuffer.CrippledHash();
                Debug.WriteLine($"c = {c.HexString()}");

                block1.XorWith(c);

                Swap(ref block1, ref block2);
                Debug.WriteLine($"a = {block1.HexString()}");
                Debug.WriteLine($"b = {block2.HexString()}");
                Debug.WriteLine("");
            }

            var returnBuffer = new byte[block.Length];
            Array.Copy(block2, 0, returnBuffer, 0, block2.Length);
            Array.Copy(block1, 0, returnBuffer, block2.Length, block1.Length);
            return returnBuffer;
        }

        private static void XorWith(this byte[] source, byte[] key)
        {
            for (var i = 0; i < source.Length; i++)
                source[i] ^= key[i];
        }

        private static void Swap<T>(ref T item1, ref T item2)
        {
            var tmp = item1;
            item1 = item2;
            item2 = tmp;
        }

        private static byte[] CrippledHash(this byte[] bytes, byte key)
        {
            var newArr = new byte[bytes.Length + 1];
            bytes.CopyTo(newArr, 0);
            newArr[^1] = key;
            return newArr.CrippledHash();
        }

        private static byte[] CrippledHash(this byte[] bytes)
        {
            using var ms = new MemoryStream(bytes);
            using var mySha256 = SHA256.Create();
            var buffer = mySha256.ComputeHash(ms);
            var newBuffer = new byte[8];
            Array.Copy(buffer, newBuffer, newBuffer.Length);
            return newBuffer;
        }

        // should replace with https://stackoverflow.com/a/9995303
        private static byte[] FromHexString(this string hex)
        {
            return Enumerable.Range(0, hex.Length)
                .Where(x => x % 2 == 0)
                .Select(x => Convert.ToByte(hex.Substring(x, 2), 16))
                .ToArray();
        }

        private static string HexString(this byte[] bytes)
        {
            return string.Join("", bytes.Select(x => x.ToString("x2")));
        }
    }
}
